#region License
// Copyright (c) 2005 Alexander Kolliopoulos
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    in a product, an acknowledgment in the product documentation would
//    be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must not
//    be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
#endregion

using System;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;

namespace Lydos.Tablets
{
    /// <summary>
    /// Provides access to tablet events and information.
    /// </summary>
    public class Tablet : IDisposable
    {
        #region Events
        /// <summary>
        /// The event that is raised when the normal pressure changes.
        /// </summary>
        public event EventHandler<TabletEventArgs> NormalPressureChange;

        /// <summary>
        /// The event that is raised when the pen moves over the tablet.
        /// </summary>
        public event EventHandler<TabletEventArgs> PenMove;

        /// <summary>
        /// The event that is raised when a button on the pen is pressed or it contacts the tablet.
        /// </summary>
        public event EventHandler<TabletEventArgs> PenPress;

        /// <summary>
        /// The event that is raised when a button on the pen is released or it is no longer touching the tablet.
        /// </summary>
        public event EventHandler<TabletEventArgs> PenRelease;

        /// <summary>
        /// The event that is raised when the pen enters the proximity of the tablet.
        /// </summary>
        public event EventHandler<EventArgs> ProximityEnter;

        /// <summary>
        /// The event that is raised when the pen leaves the proximity of the tablet.
        /// </summary>
        public event EventHandler<EventArgs> ProximityLeave;
        #endregion

        #region Fields
        Control control;
        IntPtr hTab;
        bool clientAreaEvents;
        bool consumeEvents;
        bool enabled;

        NativeMethods.Packet packet;
        NativeMethods.Packet lastPacket;

        Point screenPt;
        Point clientPt;
        #endregion

        #region Constructors
        /// <summary>
        /// Constructs a <see cref="Tablet"/> object for the specified control.
        /// </summary>
        /// <param name="control">The control the tablet is assigned to.</param>
        /// <remarks>
        /// Defaults to normal cursor behavior, so some pen events are also mouse events. The control must override its WndProc to pass messages to the tablet object.
        /// </remarks>
        public Tablet(Control control) : this(control, false) { }

        /// <summary>
        /// Construct a <see cref="Tablet"/> object for the specified control.
        /// </summary>
        /// <param name="control">The control the tablet is assigned to.</param>
        /// <param name="consumeTabletEvents">True to prevent the system from receiving tablet events, or false for normal cursor behavior.</param>
        /// <remarks>
        /// The control must override its WndProc to pass messages to the <see cref="Tablet"/> object.
        /// </remarks>
        /// <exception cref="TabletNotPresentException">Thrown when no tablets are detected.</exception>
        public Tablet(Control control, bool consumeTabletEvents)
        {
            if (!Present())
            {
                throw new TabletNotPresentException();
            }

            this.control = control;
            clientAreaEvents = true;
            consumeEvents = consumeTabletEvents;
            enabled = true;

            packet = new NativeMethods.Packet();
            lastPacket = new NativeMethods.Packet();

            screenPt = new Point(0);
            clientPt = new Point(0);

            lastPacket.Time = 0;
            lastPacket.Buttons.DW = 0;
            lastPacket.X = 0;
            lastPacket.Y = 0;
            lastPacket.NormalPressure = 0;

            screenPt = new Point(0);
            clientPt = new Point(0);

            Open();
            GC.KeepAlive(this);
        }
        #endregion

        #region Destructor
        /// <summary>
        /// Closes the connection to the tablet.
        /// </summary>
        public void Close()
        {
            if (Connected)
            {
                Dispose();
            }
        }

        /// <summary>
        /// Closes the connection to the tablet.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Final disposal of object.
        /// </summary>
        /// <param name="disposing">True if called explicitly, false if finalizing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (Connected)
            {
                NativeMethods.WTClose(hTab);
                hTab = IntPtr.Zero;
            }
        }

        /// <summary>
        /// Finalization.
        /// </summary>
        ~Tablet()
        {
            Dispose(false);
        }
        #endregion

        #region Properties
        /// <summary>
        /// The buttons currently pressed on the pen and pen contact with the tablet.
        /// </summary>
        /// <value>This comes directly from the message packet, unprocessed.</value>
        public int Buttons
        {
            get
            {
                return packet.Buttons.DW;
            }
        }

        /// <summary>
        /// Raise events when the cursor is in the client's display rectangle.
        /// </summary>
        /// <value>True to only raise events in the client area, false to raise events anywhere on the screen.</value>
        public bool ClientAreaEvents
        {
            get
            {
                return clientAreaEvents;
            }
            set
            {
                clientAreaEvents = value;
            }
        }

        /// <summary>
        /// The state of the tablet connection.
        /// </summary>
        /// <value>True it there is a valid connection to the tablet, otherwise false.</value>
        public bool Connected
        {
            get
            {
                return hTab != IntPtr.Zero;
            }
        }

        /// <summary>
        /// Consume all tablet events, or allow standard cursor behavior.
        /// </summary>
        /// <value>True if tablet messages are consumed by this context, false if the system also receives messages.</value>
        public bool ConsumeEvents
        {
            get
            {
                return consumeEvents;
            }
        }

        /// <summary>
        /// Indicates if this context is enabled or not.
        /// </summary>
        /// <value>True to enable the tablet context, false to disable it.</value>
        /// <exception cref="TabletOperationFailedException">Thrown when there is an error enabling or disabling the tablet context.</exception>
        public bool Enabled
        {
            get
            {
                return enabled;
            }
            set
            {
                if (Connected && value != enabled)
                {
                    if (value)
                    {
                        if (!NativeMethods.WTEnable(hTab, true))
                        {
                            throw new TabletOperationFailedException();
                        }
                    }
                    else
                    {
                        if (!NativeMethods.WTEnable(hTab, false))
                        {
                            throw new TabletOperationFailedException();
                        }
                    }
                    enabled = value;
                }
            }
        }

        /// <summary>
        /// Gets the name of the tablet.
        /// </summary>
        /// <value>The name of the tablet.</value>
        public string Name
        {
            get
            {
                int n = (int)NativeMethods.WTInfo((int)NativeMethods.WTI.Devices, (int)NativeMethods.DVC.Name, IntPtr.Zero);
                IntPtr namePtr = Marshal.AllocCoTaskMem(n);
                NativeMethods.WTInfo((int)NativeMethods.WTI.Devices, (int)NativeMethods.DVC.Name, namePtr);
                string name = Marshal.PtrToStringAnsi(namePtr);
                Marshal.FreeCoTaskMem(namePtr);
                return name;
            }
        }

        /// <summary>
        /// The current normal pressure for the pen on the tablet.
        /// </summary>
        /// <value>Greater values correspond to more pressure, maximum is hardware dependent.</value>
        public int NormalPressure
        {
            get
            {
                return packet.NormalPressure;
            }
        }

        /// <summary>
        /// The x-coordinate of the pen.
        /// </summary>
        /// <value>The coordinate in client space.</value>
        public int X
        {
            get
            {
                return clientPt.X;
            }
        }

        /// <summary>
        /// The y-coordinate of the pen.
        /// </summary>
        /// <value>The coordinate in client space.</value>
        public int Y
        {
            get
            {
                return clientPt.Y;
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Flushes the tablet message queue.
        /// </summary>
        public void Flush()
        {
            if (Connected)
            {
                NativeMethods.WTPacket(hTab, IntPtr.Zero, IntPtr.Zero);
                GC.KeepAlive(this);
            }
        }

        /// <summary>
        /// Gets the packet corresponding to the serial number given.
        /// </summary>
        /// <param name="wSerial">The ID of the packet to retrieve.</param>
        /// <exception cref="TabletPacketNotFoundException">Thrown when no packets are available.</exception>
        private void GetPacket(IntPtr wSerial)
        {
            int n = Marshal.SizeOf(packet);
            IntPtr pPkt = Marshal.AllocCoTaskMem(n);

            if (!NativeMethods.WTPacket(hTab, wSerial, pPkt))
            {
                throw new TabletPacketNotFoundException();
            }

            packet = (NativeMethods.Packet)Marshal.PtrToStructure(pPkt, typeof(NativeMethods.Packet));

            byte[] buf = new byte[n];
            for (int i = 0; i < n; ++i)
            {
                buf[i] = Marshal.ReadByte(pPkt, i);
            }

            Marshal.FreeCoTaskMem(pPkt);
            GC.KeepAlive(this);
        }

        /// <summary>
        /// Opens a tablet context.
        /// </summary>
        private void Open()
        {
            NativeMethods.LogContext lc;

            int n = Marshal.SizeOf(typeof(NativeMethods.LogContext));
            IntPtr context = Marshal.AllocCoTaskMem(n);

            NativeMethods.WTInfo((int)NativeMethods.WTI.DefContext, 0, context);

            lc = (NativeMethods.LogContext)Marshal.PtrToStructure(context, typeof(NativeMethods.LogContext));

            //string s = "WTDemo";
            //Encoding.ASCII.GetBytes(s, 0, s.Length, lc.Name, 0);

            lc.Options |= (ushort)NativeMethods.CXO.Messages; // Deliver messages to me
            if (!consumeEvents)  // Let the system get messages too
            {
                lc.Options |= (ushort)NativeMethods.CXO.System;
            }
            lc.MsgBase = (ushort)NativeMethods.WT.DefBase;
            lc.PktData = (uint)NativeMethods.PK.PacketData;
            lc.PktMode = 0;
            lc.MoveMask = (uint)NativeMethods.PK.PacketData;
            lc.BtnUpMask = lc.BtnDnMask;

            lc.OutOrgX = 0;
            lc.OutOrgY = 0;
            lc.OutExtX = lc.SysExtX;
            lc.OutExtY = lc.SysExtY;

            Marshal.StructureToPtr(lc, context, false);

            //byte[] buf = new byte[n];
            //for (int i = 0; i < n; ++i)
            //{
            //    buf[i] = Marshal.ReadByte(context, i);
            //}

            hTab = NativeMethods.WTOpen(control.Handle, context, true);

            Marshal.FreeCoTaskMem(context);
            GC.KeepAlive(this);
        }

        /// <summary>
        /// Processes a WndProc message and raises any appropriate events.
        /// </summary>
        /// <param name="message">The WndProc message.</param>
        public void ProcessMessage(Message message)
        {
            //if (message.Msg >= (int)NativeMethods.WT.DefBase && message.Msg <= (int)NativeMethods.WT.Max)
            //{
            //    Console.WriteLine("Incoming Message 0x{0:x} (0x{1:x8}, 0x{2:x8})", message.Msg, message.WParam, message.LParam);
            //}

            if (message.Msg == (int)NativeMethods.WT.Packet)
            {
                IntPtr s = message.WParam;

                GetPacket(s);

                screenPt.X = (int)packet.X;
                screenPt.Y = Screen.PrimaryScreen.Bounds.Height - (int)packet.Y - 1;
                clientPt = control.PointToClient(screenPt);
            }

            // Only process tablet messages that occur in the client rectangle
            if (!clientAreaEvents || control.DisplayRectangle.Contains(clientPt))
            {
                switch (message.Msg)
                {
                    case (int)NativeMethods.WT.Proximity:
                        NativeMethods.DWord lh;
                        lh.Lo = 0;
                        lh.DW = (UInt32)message.LParam;

                        if (lh.Lo > 0)
                        {
                            OnProximityEnter(EventArgs.Empty);
                        }
                        else
                        {
                            OnProximityLeave(EventArgs.Empty);
                        }
                        break;

                    case (int)NativeMethods.WT.Packet:
                        TabletEventArgs e = new TabletEventArgs(clientPt.X, clientPt.Y, packet.NormalPressure, packet.Buttons.DW);

                        if (packet.NormalPressure != lastPacket.NormalPressure)
                        {
                            OnNormalPressureChanged(e);
                        }

                        if (packet.X != lastPacket.X || packet.Y != lastPacket.Y)
                        {
                            OnPenMove(e);
                        }

                        if (packet.Buttons.DW > lastPacket.Buttons.DW)
                        {
                            OnPenPress(e);
                        }
                        else if (packet.Buttons.DW < lastPacket.Buttons.DW)
                        {
                            OnPenRelease(e);
                        }

                        lastPacket = packet;
                        break;
                }
            }
            GC.KeepAlive(this);
        }

        /// <summary>
        /// Sets the tablet context to the top or under other tablet contexts.
        /// </summary>
        /// <value>True for on top, false for under.</value>
        public void SendToTop(bool onTop)
        {
            NativeMethods.WTOverlap(hTab, onTop);
            GC.KeepAlive(this);
        }
        #endregion

        #region Event methods
        /// <summary>
        /// Raises a NormalPressureChange event.
        /// </summary>
        /// <param name="e">The event arguments.</param>
        protected virtual void OnNormalPressureChanged(TabletEventArgs e)
        {
            EventHandler<TabletEventArgs> handler = NormalPressureChange;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        /// <summary>
        /// Raises a PenMove event.
        /// </summary>
        /// <param name="e">The event arguments.</param>
        protected virtual void OnPenMove(TabletEventArgs e)
        {
            EventHandler<TabletEventArgs> handler = PenMove;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        /// <summary>
        /// Raises a PenPress event.
        /// </summary>
        /// <param name="e">The event arguments.</param>
        protected virtual void OnPenPress(TabletEventArgs e)
        {
            EventHandler<TabletEventArgs> handler = PenPress;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        /// <summary>
        /// Raises a PenRelease event.
        /// </summary>
        /// <param name="e">The event arguments.</param>
        protected virtual void OnPenRelease(TabletEventArgs e)
        {
            EventHandler<TabletEventArgs> handler = PenRelease;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        /// <summary>
        /// Raises a ProximityEnter event.
        /// </summary>
        /// <param name="e">The event arguments.</param>
        protected virtual void OnProximityEnter(EventArgs e)
        {
            EventHandler<EventArgs> handler = ProximityEnter;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        /// <summary>
        /// Raises a ProximityLeave event.
        /// </summary>
        /// <param name="e">The event arguments.</param>
        protected virtual void OnProximityLeave(EventArgs e)
        {
            EventHandler<EventArgs> handler = ProximityLeave;
            if (handler != null)
            {
                handler(this, e);
            }
        }
        #endregion

        #region Static methods
        /// <summary>
        /// Checks for the existance of a tablet.
        /// </summary>
        /// <returns>True if a tablet is available, otherwise false.</returns>
        public static bool Present()
        {
            return NativeMethods.WTInfo(0, 0, IntPtr.Zero) > 0;
        }
        #endregion
    }
}
